#include "glasslib.h"
#include "string.h"

using namespace std;

int glass_align(char* seq1, char* seq2, 
		char* alignseq1, char* alignseq2)
{

  // Convert char* to Sequence objects
  string* name1 = new string("seq1");
  string* seq1s = new string(seq1);
  Sequence* s1 = new Sequence(*name1, *seq1s);
  string* name2 = new string("seq2");
  string* seq2s = new string(seq2);
  Sequence* s2 = new Sequence(*name2, *seq2s);

  // Initialize GLASS and perform alignment
  Glass glass;
  PairwiseAlignment img = glass.align(*s1, *s2);

  // Create alignment strings

  Sequence::size_type p1 = 1;
  Sequence::size_type p2 = 1;

  Sequence::size_type hseql = s1->size();
  Sequence::size_type mseql = s2->size();

  string halign = "";
  string malign = "";

  while((p1 <= hseql) && (p2 <= mseql)) {
    if((img.firstImageGet(p1) > 0) && (img.secondImageGet(p2) > 0))  {
      halign.append(1,nucl2char(s1->get(p1)));
      malign.append(1,nucl2char(s2->get(p2)));
      p1++; p2++;
    }
    else if((img.firstImageGet(p1) == 0) && (img.secondImageGet(p2) > 0)) {
      halign.append(1,nucl2char(s1->get(p1)));
      malign.append("-");
      p1++;
    }
    else if((img.firstImageGet(p1) > 0) && (img.secondImageGet(p2) == 0)) {
      halign.append("-");
      malign.append(1,nucl2char(s2->get(p2)));
      p2++;
    }
    else if((img.firstImageGet(p1) == 0) && (img.secondImageGet(p2) == 0)) {
      if(p1 < hseql) {
	halign.append(1,nucl2char(s1->get(p1)));
	malign.append("-");
	p1++;
      }
      else {
	halign.append("-");
	malign.append(1,nucl2char(s2->get(p2)));
	p2++;
      }
    }
  }

  // Copy alignment strings into output buffers

  strcpy(alignseq1, halign.c_str()); 
  strcpy(alignseq2, malign.c_str());

  // Cleanup and return

  delete s1;
  delete name1;
  delete seq1s;
  delete s2;
  delete name2;
  delete seq2s;

  return EXIT_SUCCESS;

}


int glass_align(char* seq1, char* seq2, 
		char* alignseq1, char* alignseq2, char* alignstar)
{

  // Convert char* to Sequence objects
  string* name1 = new string("seq1");
  string* seq1s = new string(seq1);
  Sequence* s1 = new Sequence(*name1, *seq1s);
  string* name2 = new string("seq2");
  string* seq2s = new string(seq2);
  Sequence* s2 = new Sequence(*name2, *seq2s);

  // Initialize GLASS and perform alignment
  Glass glass;
  PairwiseAlignment img = glass.align(*s1, *s2);

  // Create alignment strings

  Sequence::size_type p1 = 1;
  Sequence::size_type p2 = 1;

  Sequence::size_type hseql = s1->size();
  Sequence::size_type mseql = s2->size();

  string halign = "";
  string malign = "";
  string aligns = "";

  while((p1 <= hseql) && (p2 <= mseql)) {
    if((img.firstImageGet(p1) > 0) && (img.secondImageGet(p2) > 0))  {
      halign.append(1,nucl2char(s1->get(p1)));
      malign.append(1,nucl2char(s2->get(p2)));
      if((s1->get(p1) == s2->get(p2)) && ((s1->get(p1)) != BASE_UNKNOWN)) { 
	aligns.append("*");
      }
      else { 
	aligns.append(" ");
      }
      p1++; p2++;
    }
    else if((img.firstImageGet(p1) == 0) && (img.secondImageGet(p2) > 0)) {
      halign.append(1,nucl2char(s1->get(p1)));
      malign.append("-");
      aligns.append(" ");
      p1++;
    }
    else if((img.firstImageGet(p1) > 0) && (img.secondImageGet(p2) == 0)) {
      halign.append("-");
      malign.append(1,nucl2char(s2->get(p2)));
      aligns.append(" ");
      p2++;
    }
    else if((img.firstImageGet(p1) == 0) && (img.secondImageGet(p2) == 0)) {
      if(p1 < hseql) {
	halign.append(1,nucl2char(s1->get(p1)));
	malign.append("-");
	aligns.append(" ");
	p1++;
      }
      else {
	halign.append("-");
	malign.append(1,nucl2char(s2->get(p2)));
	aligns.append(" ");
	p2++;
      }
    }
  }

  // Copy alignment strings into output buffers

  strcpy(alignseq1, halign.c_str());
  strcpy(alignseq2, malign.c_str());
  strcpy(alignstar, aligns.c_str());

  // Cleanup and return

  delete s1;
  delete name1;
  delete seq1s;
  delete s2;
  delete name2;
  delete seq2s;

  return EXIT_SUCCESS;

}

