#include "pairwisealignment.h"

// ** PairwiseAlignment methods **

PairwiseAlignment::PairwiseAlignment(const Sequence& first, 
				     const Sequence& second,
				     int* firstImg, int* secondImg) {
  firstImage.resize(first.size()+1);
  for(size_type i = 0; i <= first.size(); i++) {
    if(firstImg[i] >= 0) {
      firstImage[i] = firstImg[i];
    }
    else {
      firstImage[i] = 0;
    }
  }
  secondImage.resize(second.size()+1);
  for(size_type i = 0; i <= second.size(); i++) {
    if(secondImg[i] >= 0) {
      secondImage[i] = secondImg[i];
    }
    else {
      secondImage[i] = 0;
    }
  }
  firstSeq  = new Sequence(first);
  secondSeq = new Sequence(second);
}

PairwiseAlignment::iterator PairwiseAlignment::firstImageBegin() const 
{
  return firstImage.begin();
}

PairwiseAlignment::iterator PairwiseAlignment::firstImageEnd() const
{
  return firstImage.end();
}

Sequence::size_type PairwiseAlignment::firstImageGet(size_type pos) const 
{
  return firstImage[pos];
}

PairwiseAlignment::iterator PairwiseAlignment::secondImageBegin() const
{
  return secondImage.begin();
}

PairwiseAlignment::iterator PairwiseAlignment::secondImageEnd() const
{
  return secondImage.end();
}

Sequence::size_type PairwiseAlignment::secondImageGet(size_type pos) const 
{
  return secondImage[pos];
}

PairwiseAlignment::size_type PairwiseAlignment::firstImageSize() const {
  return firstImage.size();
}

PairwiseAlignment::size_type PairwiseAlignment::secondImageSize() const {
  return secondImage.size();
}

Sequence const* PairwiseAlignment::firstSequence() const {
  return firstSeq;
}

Sequence const* PairwiseAlignment::secondSequence() const {
  return secondSeq;
}


// ** PairwiseAlignmentIO methods **

void PairwiseAlignmentIO::writeClustal(ostream& out, 
				       const PairwiseAlignment& img) 
{
  const Sequence hseq = *(img.firstSequence());
  const Sequence mseq = *(img.secondSequence());

  Sequence::size_type p1 = 1;
  Sequence::size_type p2 = 1;

  Sequence::size_type hseql = hseq.size();
  Sequence::size_type mseql = mseq.size();

  int namel = MAX((hseq.name()).size(),(mseq.name()).size());
  int hpad  = namel - (hseq.name()).size() + 6;
  int mpad  = namel - (mseq.name()).size() + 6;
  int apad  = namel + 6;
  int alnl  = 50;
  int alnp  = 1;

  string halign = "";
  string malign = "";
  string aligns = "";

  while(true) {
    if((p1 > hseql) || (p2 > mseql)) {
      // Finish up remaining sequence. Only one of these loops run. 
      while(p1 <= hseql) {
	if(alnp == 1) {
	  halign = hseq.name();
	  for(int i = 0; i < hpad; i++) {
	    halign.append(" ");
	  }
	  malign = mseq.name();
	  for(int i = 0; i < mpad; i++) {
	    malign.append(" ");
	  }
	  aligns = "";
	  for(int i = 0; i < apad; i++) {
	    aligns.append(" ");
	  }
	} 
	else if(alnp == alnl) {
	  out << halign << "\n" << malign << "\n" << aligns << "\n" << endl;
	  alnp = 0;
	}
	else {
	  halign.append(1,nucl2char(hseq.get(p1)));
	  malign.append("-");
	  aligns.append(" ");
	  p1++;
	}
	++alnp;    
      }
      while(p2 <= mseql) {
	if(alnp == 1) {
	  halign = hseq.name();
	  for(int i = 0; i < hpad; i++) {
	    halign.append(" ");
	  }
	  malign = mseq.name();
	  for(int i = 0; i < mpad; i++) {
	    malign.append(" ");
	  }
	  aligns = "";
	  for(int i = 0; i < apad; i++) {
	    aligns.append(" ");
	  }
	} 
	else if(alnp == alnl) {
	  out << halign << "\n" << malign << "\n" << aligns << "\n" << endl;
	  alnp = 0;
	}
	else {
	  halign.append("-");
	  malign.append(1,nucl2char(mseq.get(p2)));
	  aligns.append(" ");
	  p2++;
	}
	++alnp;    
      }
      // Print our remaining sequence.
      if(alnp != 1) {
	out << halign << "\n" << malign << "\n" << aligns << "\n" << endl;
      }
      break;
    }

    if(alnp == 1) {
      halign = hseq.name();
      for(int i = 0; i < hpad; i++) {
	halign.append(" ");
      }
      malign = mseq.name();
      for(int i = 0; i < mpad; i++) {
	malign.append(" ");
      }
      aligns = "";
      for(int i = 0; i < apad; i++) {
	aligns.append(" ");
      }
    } 
    else if(alnp == alnl) {
      out << halign << "\n" << malign << "\n" << aligns << "\n" << endl;
      alnp = 0;
    }
    else if((img.firstImageGet(p1) > 0) && (img.secondImageGet(p2) > 0))  {
      halign.append(1,nucl2char(hseq.get(p1)));
      malign.append(1,nucl2char(mseq.get(p2)));
      if((hseq.get(p1) == mseq.get(p2)) && ((hseq.get(p1)) != BASE_UNKNOWN)) { 
	aligns.append("*");
      }
      else { 
	aligns.append(" ");
      }
      p1++; p2++;
    }
    else if((img.firstImageGet(p1) == 0) && (img.secondImageGet(p2) > 0)) {
      halign.append(1,nucl2char(hseq.get(p1)));
      malign.append("-");
      aligns.append(" ");
      p1++;
    }
    else if((img.firstImageGet(p1) > 0) && (img.secondImageGet(p2) == 0)) {
      halign.append("-");
      malign.append(1,nucl2char(mseq.get(p2)));
      aligns.append(" ");
      p2++;
    }
    else if((img.firstImageGet(p1) == 0) && (img.secondImageGet(p2) == 0)) {
      if(p1 < hseql) {
	halign.append(1,nucl2char(hseq.get(p1)));
	malign.append("-");
	aligns.append(" ");
	p1++;
      }
      else {
	halign.append("-");
	malign.append(1,nucl2char(mseq.get(p2)));
	aligns.append(" ");
	p2++;
      }
    }
    if((p1 <= hseql) && (p2 <= mseql)) {
      assert(img.firstImageGet(p1) >= 0 || img.secondImageGet(p2) >= 0);
    }

    ++alnp;
  }
}


void PairwiseAlignmentIO::writeRegions(ostream& out, 
				       const vector<PairwiseAlignment>& img) 
{
  for(int i = 0; i < img.size(); i++) {
    out << "Region " << (i+1) << '/' << img.size() << "\n\n";
    PairwiseAlignmentIO::writeClustal(out, img[i]);
  }
}
