#ifndef __PAIRWISEALIGNMENT_H__
#define __PAIRWISEALIGNMENT_H__

#include "seq.h"
#include "common.h"

// 1-indexed immutable pairwise alignment class

class PairwiseAlignment
{
 public:

  typedef vector<Sequence::size_type>                 image_type;
  typedef vector<Sequence::size_type>::size_type      size_type;
  typedef vector<Sequence::size_type>::const_iterator iterator;

  PairwiseAlignment(const Sequence& first, const Sequence& second,
		    int* firstImg, int* secondImg);
  //
  // Requires:  first.size()  == length(firstImg)  &&
  //            second.size() == length(secondImg)
  //
  // Effects:   Creates a new Pairwisealignment 
  //

  iterator firstImageBegin() const;
  iterator firstImageEnd()   const;
  Sequence::size_type firstImageGet(size_type pos) const;

  iterator secondImageBegin() const;
  iterator secondImageEnd()   const;
  Sequence::size_type secondImageGet(size_type pos) const;

  Sequence const* firstSequence()  const;
  Sequence const* secondSequence() const;

  size_type firstImageSize()  const;
  size_type secondImageSize() const;

 protected:
 
  Sequence* firstSeq;
  Sequence* secondSeq;

  image_type firstImage;
  image_type secondImage;

};

// Method collection for PairwiseAlignment I/O

class PairwiseAlignmentIO 
{

 public:

  static void writeClustal(ostream& out, const PairwiseAlignment& img) ;
  // modifies: out
  // effects:  Writes img to out in Clustal W format.

  static void writeRegions(ostream& out, const vector<PairwiseAlignment>& img);
  // modifies: out
  // effects:  Writes each alignment in imgs to out in order. The ith alignment
  //           is preceded by the header 'Region i/n', where n is the number 
  //           of elements in imgs. Each Alignment is written in Clustal W
  //           format.

};

#endif
