#include "seq.h"

Sequence::Sequence(string &name, string &s) 
{
  seqname = name;
  for(int i = 0; i < s.size(); i++) {
    seq.push_back(char2nucl(s[i]));
  }
}

Sequence::Sequence(const Sequence& s) {
  seqname = s.seqname;
  seq.insert(seq.end(),(s.seq).begin(), (s.seq).end());
}

Sequence::Sequence(const Sequence& s, size_type from, size_type to) {
  char subs[256];
  sprintf(subs,"(%lu,%lu)",from,to);
  seqname = s.name();
  seqname.append(subs);
  seq.resize(to - from + 1);
  for(int i = 0; i < seq.size(); i++) {
    seq[i] = s.seq[from + i - 1];
  }
}

Sequence::~Sequence() 
{
}

string Sequence::name() const
{
  return seqname;
}

vector<Nucleotide>::size_type Sequence::size() const
{
  return seq.size();
}

vector<Nucleotide>::iterator Sequence::begin()
{
  return seq.begin();
}

vector<Nucleotide>::iterator Sequence::end()
{
  return seq.end();
}

Nucleotide& Sequence::operator[](size_type pos) 
{  
  return seq[pos-1]; 
}                      

const Nucleotide& Sequence::get(size_type pos) const {
  return seq[pos-1];
}

ostream& operator<<(ostream& out, Sequence& seq) 
{
  for ( int i = 0;
	i < seq.size();
	++i )
    out << nucl2char( seq.seq[ i ] );

  return out;
}

int seq2int(Sequence* seq, int* seqInt) {
  seqInt[0] = 0;
  int seql = seq->size();
  int i;
  for(i = 0; i < seql; i++) {
    seqInt[i+1] = (int)seq->seq[i];
  }
  return i;
}

Nucleotide char2nucl(char ch) {
  switch (ch) {
  case 'a':
  case 'A':
    return BASE_A;
  case 'c':
  case 'C':
    return BASE_C;
  case 'g':
  case 'G':
    return BASE_G;
  case 't':
  case 'T':
    return BASE_T;
  default:
    return BASE_UNKNOWN;
  }
}

char nucl2char(Nucleotide base) {
  switch (base) {
    case BASE_A:
      return 'a';
    case BASE_C:
      return 'c';
    case BASE_G:
      return 'g';
    case BASE_T:
      return 't';
  default:
    return 'n';
  }
}


// ** SequenceIO methods **

Sequence SequenceIO::readFasta(istream& in)
{
  string name = "noname";
  string seq = "";
  char c = in.get();  // Remove '>'
  if(c == '>') {
    char buf[1000];
    in.getline( buf, 1000 );
    buf[ 999 ] = '\0';
    istrstream namestr( buf );
    namestr >> name;
    if(name.size() >= 256) {
      name = name.substr(0,256);
    }
  }
  else {
    in.putback(c);
  }
  string n;
  while(in >> n) 
    {
      seq.append(n);
    }
  Sequence s = Sequence(name, seq);
  return s;
}

void SequenceIO::writeFasta(ostream& out, Sequence& s) 
{
  out << '>' << s.name() << endl;
  int pos = 1;
  for(Sequence::iterator i = s.begin(); i != s.end(); i++)
    {
      out << nucl2char(*i);
      if((pos % 60) == 0)
	{
	  out << '\n';
	  pos = 1;
	}
      else
	{
	  pos++;
	}
    }
  out << endl;
}








