#ifndef __SEQ_H__
#define __SEQ_H__

#include <cstdio>
#include <iostream>
#include <strstream>
#include <string>
#include <vector>

using namespace std;

// Binary encoding of the nucleotides
enum Nucleotide 
{
  BASE_A = 0, 
  BASE_C = 1, 
  BASE_G = 2, 
  BASE_T = 3, 
  BASE_UNKNOWN
};

// 1-indexed DNA sequence class

class Sequence 
{
 public:

  typedef vector<Nucleotide>::size_type size_type;
  typedef vector<Nucleotide>::iterator  iterator;

  Sequence(string &name, string &s);
  Sequence(const Sequence& s);

  Sequence(const Sequence& s, size_type from, size_type to);
  // 
  // Requires: s.size() < to
  //
  // Effects:  Constructs a new sequence s[from..to] (inclusive) 
  //           whose name is s.name() + "(from,to)" 
  //

  virtual ~Sequence();

  string name() const;
  size_type size() const;

  iterator begin();
  iterator end();
  Nucleotide& operator[](size_type pos);
  const Nucleotide& get(size_type pos) const;

  friend ostream& operator<<(ostream& out, Sequence& seq);

 private:

  string seqname;
  vector<Nucleotide> seq;

  friend int seq2int(Sequence*, int*);  
};

int seq2int(Sequence* seq, int *seqInt);
// Does NOT allocate memory for seqInt;
// Returns sequence length;
// seqInt[0] is arbitrary (in fact it is 0);


Nucleotide char2nucl(char ch);

char nucl2char(Nucleotide base);


class SequenceIO 
{
 public:
  static void writeFasta(ostream& out, Sequence& seq);
  // modifies: out
  // effects:  Writes seq to out in 'fasta' format:
  //           >name
  //           agtcagtc....\n
  //           agtcagtc....\n

  static Sequence readFasta(istream& in);
  // requires: in contains only a single sequence.
  // modifies: in
  // effects:  Reads in a sequence in 'fasta' format.
  //           The '>' header is optional. If it is not
  //           supplied, the output Sequence will be
  //           named 'noname'.

};

#endif // __SEQ_H__




















